#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Dict, List

from mesh.macro import spi
from mesh.prsim import Hodor


@spi("mesh")
class MeshHodor(Hodor):

    def stats(self, features: List[str]) -> Dict[str, str]:
        return {"status": "true"}

    def debug(self, features: Dict[str, str]):
        """ debug """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.ioc.context import Context
from mesh.ioc.envs_processor import EnvsProcessor
from mesh.ioc.paas_processor import PaaSProcessor

__all__ = ("PaaSProcessor", "EnvsProcessor", "Context")


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import ABC, abstractmethod
from typing import List

from mesh.kinds import License, CommerceLicense, CommerceEnviron
from mesh.macro import mpi


class Commercialize(ABC):

    @abstractmethod
    @mpi(name="mesh.license.sign", flags=2)
    def sign(self, lsr: License) -> str:
        """
        Sign the license.
        :param lsr:
        :return:
        """
        pass

    @abstractmethod
    @mpi(name="mesh.license.history", flags=2)
    def history(self, inst_id: str) -> List[CommerceLicense]:
        """
        History list the licenses.
        :param inst_id:
        :return:
        """
        pass

    @abstractmethod
    @mpi(name="mesh.net.issued", flags=2)
    def issued(self, name: str, kind: str) -> CommerceEnviron:
        """
        Issued mesh node identity.
        :param name:
        :param kind:
        :return:
        """
        pass

    @abstractmethod
    @mpi(name="mesh.net.dump", flags=2)
    def issued(self, node_id: str) -> List[CommerceEnviron]:
        """
        Dump the node identity.
        :param node_id:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import time
from typing import Any, Dict, Type

import mesh.log as log
import mesh.telemetry as telemetry
import mesh.tool as tool
from mesh.cause import MeshCode, Codeable, MeshException
from mesh.codec import Codec
from mesh.context import Mesh, MeshContext
from mesh.kinds import MeshFlag, Location
from mesh.macro import spi, ServiceLoader
from mesh.mpc.digest import Digest
from mesh.mpc.filter import Filter, Invoker, Invocation, PROVIDER
from mesh.prsim import Metadata
from mesh.prsim.context import RunMode


@spi(name="provider", pattern=PROVIDER, priority=0x7fffff00)
class ProviderFilter(Filter):

    def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        codec: Codec = ServiceLoader.load(Codec).get(MeshFlag.JSON.get_name())
        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        if not attachments:
            attachments = {}
        trace_id = attachments.get(Metadata.MESH_TRACE_ID.key(), tool.new_trace_id())
        span_id = attachments.get(Metadata.MESH_SPAN_ID.key(), tool.new_span_id("", 0))
        timestamp = attachments.get(Metadata.MESH_TIMESTAMP.key(), '')
        run_mode = attachments.get(Metadata.MESH_RUN_MODE.key(), '')
        urn = attachments.get(Metadata.MESH_URN.key())
        provider = attachments.get(Metadata.MESH_PROVIDER.key(), '{}')
        pp: Location = codec.decode_string(provider, Type[Location])

        # Refresh context
        context = MeshContext()
        context.trace_id = trace_id
        context.span_id = span_id
        context.timestamp = ProviderFilter.resolve_timestamp(timestamp)
        context.run_mode = ProviderFilter.resolve_run_mode(run_mode)
        context.urn = urn
        context.consumer = pp
        for key, value in attachments.items():
            context.attachments[key] = value
        Mesh.context().rewrite_context(context)

        digest = Digest()
        try:
            ret = invoker.run(invocation)
            digest.write("P", MeshCode.SUCCESS.get_code())
            return ret
        except BaseException as e:
            if isinstance(e, Codeable):
                digest.write("P", e.get_code())
            else:
                digest.write("P", MeshCode.SYSTEM_ERROR.get_code())
            if isinstance(e, MeshException):
                log.error(f"{digest.trace_id},{Mesh.context().get_urn()},{e.get_message}")
            raise e

    @staticmethod
    def resolve_timestamp(v: str) -> int:
        if tool.optional(v):
            return int(time.time() * 1000)
        try:
            return int(v)
        except Exception as e:
            log.error("Parse timestamp failed. ", e)
            return int(time.time() * 1000)

    @staticmethod
    def resolve_run_mode(v: str) -> RunMode:
        if tool.optional(v):
            return RunMode.ROUTINE
        try:
            return RunMode.get_by_code(int(v))
        except Exception as e:
            log.error("Parse run mode failed.", e)
            return RunMode.ROUTINE


@spi(name="telemetryProvider", pattern=PROVIDER, priority=0x7ffffff0)
class TelemetryProviderFilter(Filter):
    """
    recover telemetry context via mesh context for tracing
    """

    def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        if not telemetry.if_rebuild_span():
            return invoker.run(invocation)
        trace_id = attachments.get(Metadata.MESH_TRACE_ID.key(), tool.new_trace_id())
        with telemetry.build_via_remote(attachments, invocation.get_urn().string()):
            current_span = telemetry.get_current_span()
            current_span.set_attribute('mesh-trace-id', trace_id)
            return invoker.run(invocation)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from __future__ import absolute_import, division, unicode_literals

from mesh.context.context import MeshKey, MeshContext, Queue
from mesh.context.mesh import Mesh

__all__ = (
    "Mesh",
    "MeshContext",
    "MeshKey",
    "Queue",
)


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import random
import time
from typing import Dict, Any, Optional, Type

import mesh.tool as tool
from mesh.kinds import Principal, Location
from mesh.macro import ServiceLoader
from mesh.macro import T
from mesh.prsim import Context
from mesh.prsim import Network
from mesh.prsim.context import Key, RunMode, Queue


class MeshContext(Context):

    @staticmethod
    def create() -> Context:
        return MeshContext()

    def __init__(self):
        self.trace_id = tool.new_trace_id()
        self.span_id = tool.new_span_id("", 0)
        self.timestamp = int(time.time() * 1000)
        self.run_mode = RunMode.ROUTINE
        self.urn = ''
        self.calls = 0
        self.consumer = Location()
        self.attachments = dict()
        self.attachments['mesh-telemetry-trace-id'] = hex(random.getrandbits(128))[2:]
        self.attributes = dict()
        self.principals = Queue()

    def get_trace_id(self) -> str:
        return self.trace_id

    def get_span_id(self) -> str:
        return self.span_id

    def get_timestamp(self) -> int:
        return self.timestamp

    def get_run_mode(self) -> RunMode:
        return self.run_mode

    def get_urn(self) -> str:
        return self.urn

    def get_consumer(self) -> Location:
        if self.consumer and tool.required(self.consumer.node_id):
            return self.consumer
        return self.get_provider()

    def get_provider(self) -> Location:
        network = ServiceLoader.load(Network).get_default()
        environ = network.get_environ()
        location = Location()
        location.inst_id = environ.inst_id
        location.node_id = environ.node_id
        location.ip = tool.get_mesh_direct()
        location.host = tool.get_hostname()
        location.port = f"{tool.get_mesh_runtime().port}"
        location.name = tool.get_mesh_name()
        return location

    def get_attachments(self) -> Dict[str, str]:
        return self.attachments

    def get_principals(self) -> Queue[Principal]:
        return self.principals

    def get_attributes(self) -> Dict[str, Any]:
        return self.attributes

    def get_attribute(self, key: Key[T]) -> T:
        return self.attributes.get(key.name, None)

    def set_attribute(self, key: Key[T], value: T) -> None:
        self.attributes[key.name] = value

    def rewrite_urn(self, urn: str) -> None:
        self.urn = urn

    def rewrite_context(self, another: Context) -> None:
        if tool.required(another.get_trace_id()):
            self.trace_id = another.get_trace_id()
        if tool.required(another.get_span_id()):
            self.span_id = another.get_span_id()
        if another.get_timestamp() > 0:
            self.timestamp = another.get_timestamp()
        if RunMode.ROUTINE != another.get_run_mode():
            self.run_mode = another.get_run_mode()
        if tool.required(another.get_urn()):
            self.urn = another.get_urn()
        if tool.required(another.get_consumer()):
            self.consumer = another.get_consumer()
        if tool.required(another.get_attachments()):
            for key, value in another.get_attachments().items():
                self.attachments[key] = value
        if 'mesh-telemetry-trace-id' not in another.get_attachments():
            self.attachments['mesh-telemetry-trace-id'] = hex(random.getrandbits(128))[2:]
        if tool.required(another.get_attributes()):
            for key, value in another.get_attributes().items():
                self.attributes[key] = value
        if tool.required(another.get_principals()):
            for value in another.get_principals():
                self.principals.append(value)

    def resume(self) -> Context:
        self.calls = self.calls + 1
        context = MeshContext()
        context.rewrite_context(self)
        return context


class MeshKey(Key):

    def __init__(self, name: str, kind: Type[T]):
        self.name = name
        self.kind = kind

    def get_if_absent(self) -> T:
        pass

    def map(self, fn) -> Optional[Any]:
        pass

    def if_present(self, fn):
        pass

    def or_else(self, v: T) -> T:
        pass

    def is_present(self) -> bool:
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import random

from opentelemetry import trace
from opentelemetry.trace import Tracer, NonRecordingSpan
from opentelemetry.trace.span import Span, SpanContext, TraceFlags, INVALID_SPAN

from mesh.environ import System
from mesh.telemetry.telemetry import MeshTelemetry

mot = MeshTelemetry()


def init():
    """ init function """
    pass


def provider_tracer(tracer_name: str = '', print_std: bool = False) -> "Tracer":
    """
    :param tracer_name: app name may be better
    :param print_std: True: print metrics data on stdout; False: ignore
    :return: Tracer, used as below,
    for example:
    with tracer.start_as_current_span("asm.init"):
        pass
    """

    __tracer__ = mot.get_tracer(enable_std=print_std)
    if __tracer__:
        if tracer_name != '':
            return __tracer__.get_tracer(tracer_name)
        else:
            return __tracer__.get_tracer(System.environ().get_mesh_name())
    else:
        print('open telemetry not active!')
        # raise Exception('open telemetry not active!')


def get_current_span() -> Span:
    return trace.get_current_span()


tracer = provider_tracer(print_std=System.environ().enable_telemetry_std())


def if_rebuild_span() -> bool:
    # tracer may be not active
    if not tracer:
        return False
    return True


def build_via_remote(attachments: dict, name: str):
    """
    enable only for if_rebuild_span return True
    :param name:
    :param attachments: mesh context
    :return: telemetry Span
    """
    ctx = None
    trace_id = 0
    span_id = random.getrandbits(64)
    if 'mesh-telemetry-trace-id' in attachments:
        trace_id = int(attachments['mesh-telemetry-trace-id'], 16)
    if trace.get_current_span() == INVALID_SPAN or \
            (trace_id != 0 and get_current_span().get_span_context().trace_id != trace_id):
        span_context = SpanContext(trace_id=_choose_trace_id(trace_id),
                                   span_id=span_id,
                                   is_remote=True,
                                   trace_flags=TraceFlags.SAMPLED)
        ctx = trace.set_span_in_context(NonRecordingSpan(span_context))
    return tracer.start_as_current_span(name=name, context=ctx)


def _choose_trace_id(trace_id: int) -> int:
    if trace_id == 0:
        return random.getrandbits(128)
    return trace_id



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from opentelemetry import trace
from opentelemetry.sdk.trace import (
    TracerProvider,
)
from opentelemetry.sdk.trace.export import (
    ConsoleSpanExporter,
    SimpleSpanProcessor,
)
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace.sampling import ALWAYS_ON

from mesh.environ import System


class MeshTelemetry:

    def __init__(self):
        self.__tracer__ = None

    def get_tracer(self, service_name: str = 'gaia-tracer', enable_std: bool = False):
        if not System.environ().enable_telemetry():
            return None
        if self.__tracer__:
            return self.__tracer__
        # Service name is required for most backends
        resource = Resource(attributes={
            SERVICE_NAME: service_name
        })
        provider = TracerProvider(resource=resource, sampler=ALWAYS_ON)

        if enable_std:
            std_processor = SimpleSpanProcessor(ConsoleSpanExporter())
            provider.add_span_processor(std_processor)

        if System.environ().get_telemetry_protocol() == 'http':
            from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
            span_processor = SimpleSpanProcessor(OTLPSpanExporter(endpoint=System.environ().get_telemetry_endpoint()))
            provider.add_span_processor(span_processor)
        else:
            from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
            span_processor = SimpleSpanProcessor(OTLPSpanExporter(endpoint=System.environ().get_telemetry_endpoint(),
                                                                  insecure=True))
            provider.add_span_processor(span_processor)

        trace.set_tracer_provider(provider)
        self.__tracer__ = provider
        return self.__tracer__



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import multiprocessing
from concurrent import futures
from typing import Any

import grpc

import mesh.log as log
import mesh.tool as tool
from mesh.cause import ValidationException
from mesh.grpx import GrpcBindableService, GrpcInterceptor
from mesh.macro import spi
from mesh.mpc import Provider


@spi("grpc")
class GrpcProvider(Provider):
    """
    Grpc provider wrap the grpc protocol implementation.
    """

    def __init__(self):
        self.interceptor = GrpcInterceptor()
        self.service = GrpcBindableService()
        self.server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=multiprocessing.cpu_count() * 10),
            [self.service],
            [self.interceptor],
            [('grpc.max_message_length', 1 << 29)],
            None,
            grpc.Compression.Gzip,
            False
        )

    def start(self, address: str, tc: Any):
        if tool.optional(address):
            raise ValidationException("GRPC address cant be empty.")
        self.server.add_insecure_port(address)
        self.server.add_generic_rpc_handlers([self.service])
        self.server.start()
        log.info(f"Listening and serving HTTP 2.0 on {address}")

    def close(self):
        self.server.stop(grace=12)
        log.info(f"Graceful stop HTTP 2.0 serving")



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#


from typing import Any, Type

from mesh.codec import Codec, Json
from mesh.macro import index, ServiceLoader, T, serializable


@serializable
class Entity:

    @staticmethod
    def wrap(value: Any) -> "Entity":
        if value is None:
            return Entity()
        cdc = ServiceLoader.load(Codec)
        entity = Entity()
        entity.codec = cdc.default_name()
        entity.schema = ""
        entity.buffer = cdc.get_default().encode(value)
        return entity

    def present(self) -> bool:
        return self.schema is not None and self.schema != ""

    def read_object(self) -> T:
        return

    def try_read_object(self, kind: Type[T]) -> T:
        return self.load_codec().decode(self.buffer, kind) if self.buffer else None

    def load_codec(self) -> Codec:
        if not self.codec or self.codec == "":
            return ServiceLoader.load(Codec).get(Json)
        return ServiceLoader.load(Codec).get(self.codec)

    @index(0)
    def codec(self) -> str:
        return ""

    @index(5)
    def schema(self) -> str:
        return ""

    @index(10)
    def buffer(self) -> bytes:
        return bytes()


@serializable
class CacheEntity:

    @index(0)
    def version(self) -> str:
        return ""

    @index(5)
    def entity(self) -> Entity:
        return Entity()

    @index(10)
    def timestamp(self) -> int:
        return 0

    @index(15)
    def duration(self) -> int:
        return 0

    @index(20)
    def key(self) -> str:
        return ""



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.macro import index, serializable


@serializable
class Route:

    @index(0)
    def node_id(self) -> str:
        """ 节点编号 """
        return ""

    @index(5)
    def inst_id(self) -> str:
        """ 机构编号 """
        return ""

    @index(10)
    def name(self) -> str:
        """ 节点名称 """
        return ""

    @index(15)
    def inst_name(self) -> str:
        """ 机构编号 """
        return ""

    @index(20)
    def address(self) -> str:
        """ 节点地址 """
        return ""

    @index(25)
    def describe(self) -> str:
        """ 节点描述 """
        return ""

    @index(30)
    def host_root(self) -> str:
        """ Host root certifications """
        return ""

    @index(35)
    def host_key(self) -> str:
        """ Host private certifications key """
        return ""

    @index(40)
    def host_crt(self) -> str:
        """ Host certification """
        return ""

    @index(45)
    def guest_root(self) -> str:
        """ Guest root certifications """
        return ""

    @index(50)
    def guest_key(self) -> str:
        """ Guest private certifications key """
        return ""

    @index(55)
    def guest_crt(self) -> str:
        """ Guest certifications key """
        return ""

    @index(60)
    def status(self) -> int:
        """ 状态 """
        return 0

    @index(65)
    def version(self) -> int:
        """ 状态 """
        return 0

    @index(70)
    def auth_code(self) -> str:
        """ 授权码 """
        return ""

    @index(75)
    def expire_at(self) -> int:
        """ 失效时间 """
        return 0

    @index(80)
    def extra(self) -> str:
        """ 扩展信息 """
        return ""

    @index(85)
    def create_at(self) -> str:
        """ 创建时间 """
        return ""

    @index(90)
    def create_by(self) -> str:
        """ 创建人 """
        return ""

    @index(95)
    def update_at(self) -> str:
        """ 更新时间 """
        return ""

    @index(100)
    def update_by(self) -> str:
        """ 跟新人 """
        return ""

    @index(105)
    def group(self) -> str:
        """ Network group. """
        return ""

    @index(110)
    def upstream(self) -> int:
        """ Upstream bandwidth. """
        return 0

    @index(115)
    def downstream(self) -> int:
        """ Downstream bandwidth. """
        return 0

    @index(120)
    def static_ip(self) -> str:
        """ Static public ip address """
        return ""

    @index(125)
    def proxy(self) -> str:
        """ Proxy endpoint in transport """
        return ""

    @index(130)
    def concurrency(self) -> int:
        """ MPC concurrency """
        return 0



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import inspect
from abc import ABCMeta, abstractmethod


class Interface(metaclass=ABCMeta):
    pass


def interface(cls):
    attrs = {n: abstractmethod(f)
             for n, f in inspect.getmembers(cls, predicate=inspect.isfunction)}

    return type(cls.__name__, (Interface, cls), attrs)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import unittest

import runtime


class TestMacro(unittest.TestCase):

    def test_start_container(self):
        container = runtime.Container()
        container.start()

    def test_stop_container(self):
        container = runtime.Container()
        container.start()
        container.stop()


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from enum import Enum


class MeshFlag(Enum):
    HTTP = "00", "http", 0
    GRPC = "01", "grpc", 0
    MQTT = "02", "mqtt", 0
    TCP = "03", "tcp", 0

    JSON = "00", "json", 1
    PROTOBUF = "01", "protobuf", 1
    XML = "02", "xml", 1
    THRIFT = "03", "thrift", 1
    YAML = "04", "yaml", 1

    def get_code(self):
        return self.value[0]

    def get_name(self):
        return self.value[1]

    @staticmethod
    def of_proto(code: str) -> "MeshFlag":
        for member in MeshFlag:
            if member.value and member.value[2] == 0 and member.value[0] == code:
                return member
        return MeshFlag.HTTP

    @staticmethod
    def of_code(code: str) -> "MeshFlag":
        for member in MeshFlag:
            if member.value and member.value[2] == 1 and member.value[0] == code:
                return member
        return MeshFlag.JSON

    @staticmethod
    def of_name(name: str) -> "MeshFlag":
        for member in MeshFlag:
            if member.value and member.value[1] == name:
                return member
        return MeshFlag.HTTP



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.kinds import CacheEntity
from mesh.macro import spi
from mesh.mpc import ServiceProxy
from mesh.prsim import Cache


@spi("mesh")
class MeshCache(Cache):

    def __init__(self):
        self.proxy = ServiceProxy.default_proxy(Cache)

    def get(self, key: str) -> CacheEntity:
        return self.proxy.get(key)

    def put(self, cell: CacheEntity) -> None:
        return self.proxy.put(cell)

    def remove(self, key: str):
        return self.proxy.remove(key)

    def incr(self, key: str, value: int, duration: int) -> int:
        return self.proxy.incr(key, value, duration)

    def decr(self, key: str, value: int, duration: int) -> int:
        return self.proxy.decr(key, value, duration)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#


from mesh.runtime.container import Container
from mesh.runtime.context import MeshContext

__all__ = ("Container", "MeshContext")


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod

from mesh.macro import spi, mpi


@spi("mesh")
class Locker(ABC):

    @abstractmethod
    @mpi("mesh.locker.w.lock")
    def lock(self, rid: str, timeout: int) -> bool:
        """
        Acquires the lock.
        :param rid:
        :param timeout: in mills
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.locker.w.unlock")
    def unlock(self, rid: str):
        """
        Releases the lock.
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.locker.r.lock")
    def read_lock(self, rid: str, timeout: int) -> bool:
        """
        Create a sample lock.
        :param rid:
        :param timeout:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.locker.r.unlock")
    def read_unlock(self, rid: str):
        """
        Create a sample lock.
        :param rid:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Generic

from mesh.macro import T, index as idx, serializable


@serializable
class Page(Generic[T]):

    @idx(0)
    def sid(self) -> str:
        return ''

    @idx(5)
    def index(self) -> int:
        return 0

    @idx(10)
    def limit(self) -> int:
        return 0

    @idx(15)
    def total(self) -> int:
        return 0

    @idx(20)
    def next(self) -> bool:
        return False

    @idx(25)
    def data(self) -> T:
        return None



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import ABC, abstractmethod
from typing import List

from mesh.macro import spi, mpi


@spi("mesh")
class Sequence(ABC):

    @abstractmethod
    @mpi("mesh.sequence.next")
    def next(self, kind: str, length: int) -> str:
        """
        Generate a unique number in network.
        :param kind:
        :param length:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.sequence.section")
    def section(self, kind: str, size: int, length: int) -> List[str]:
        """
        Generate some unique number in network as s section.
        :param kind:
        :param size:
        :param length:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any

from mesh.ioc import Context, EnvsProcessor, PaaSProcessor
from mesh.kinds.reference import Reference
from mesh.kinds.service import Service


class MeshContext(Context):

    def __init__(self):
        self.env_processors = [EnvsProcessor]
        self.paas_processors = [PaaSProcessor]
        self.properties = []
        self.services = [Service]
        self.references = [Reference]

    def inject_properties(self, properties: bytes):
        self.properties.append(bytes)

    def register_processor(self, processor: Any):
        if issubclass(processor.__class__, EnvsProcessor.__class__):
            self.env_processors.append(processor)
        if issubclass(processor.__class__, PaaSProcessor.__class__):
            self.paas_processors.append(processor)

    def refresh(self):
        services = []
        for processor in self.env_processors:
            processor.post_properties(self)
        for service in self.services:
            for processor in self.paas_processors:
                service = processor.before_initialization(service.kind, service.name)
                service = processor.after_initialization(service, service.name)
                service = processor.before_instantiation(service, service.name)
                service = processor.after_instantiation(service, service.name)
                services = services.append(service)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Dict, Any, List

from mesh.kinds import Script, Paging, Page
from mesh.macro import spi
from mesh.mpc import ServiceProxy
from mesh.prsim import Evaluator


@spi("mesh")
class MeshEvaluator(Evaluator):

    def __init__(self):
        self.evaluator = ServiceProxy.default_proxy(Evaluator)

    def compile(self, script: Script) -> str:
        return self.evaluator.compile(script)

    def exec(self, code: str, args: Any, dft: str) -> str:
        return self.evaluator.exec(code, args, dft)

    def dump(self, feature: Dict[str, str]) -> List[Script]:
        return self.evaluator.dump(feature)

    def index(self, index: Paging) -> Page[Script]:
        return self.evaluator.index(index)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import abstractmethod, ABC

from mesh.macro import spi, mpi


@spi(name="mesh")
class Cluster(ABC):

    @abstractmethod
    @mpi("mesh.cluster.election")
    def election(self, buff: bytes) -> bytes:
        """
        Election will election leader of instances.
        :param buff:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.cluster.leader")
    def is_leader(self) -> bool:
        """
        IsLeader if same level.
        :return:
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.macro import index, serializable


@serializable
class Institution:

    @index(0)
    def node_id(self) -> str:
        return ''

    @index(5)
    def inst_id(self) -> str:
        return ''

    @index(10)
    def inst_name(self) -> str:
        return ''

    @index(15)
    def status(self) -> int:
        return 0
